/*! ********************************************
 * EFXStereoDelay.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.efx
{
import sora.graybox.GrayBox;
import sora.graybox.ProcessData;
import sora.graybox.IGrayBoxModule;
public class EFXStereoDelay extends EffectorBase implements IGrayBoxModule
{
// EFXStereoDelay --- 
	// Ring Buffer
	private var _ringBufferSizeBits:int = 15;
	private var _ringBufferSize:int = 1 << _ringBufferSizeBits;
	private var _ringBufferMask:int = _ringBufferSize - 1;
	private var _delayBufferL:Vector.<Number>;
	private var _delayBufferR:Vector.<Number>;
	// 
	private var _sampleRate:Number = 44100.0;
	// Delay
	private var _wp :int = 0;  // WritePoint [frame]
	// Delay L
	private var _feedBackL  :Number = 0.4; // 0 ~ 1.0
	private var _rpL        :int = 0;      // ReadPoint [frame]
	private var _delayL     :int = 0;      // [frame]
	private var _delayTimeL :Number = 0.0; // [Sec]
	// Delay R
	private var _feedBackR  :Number = 0.4;
	private var _rpR        :int = 0;
	private var _delayR     :int = 0;
	private var _delayTimeR :Number = 0.0;
// getter, setter
	public function getRingBufferSize():int { return _ringBufferSize; }
	public function getSampleRate():Number { return _sampleRate; }
	public function setSampleRate(rate:Number):void
	{
		_sampleRate = rate;
		_delayTimeL = _delayL / _sampleRate;
		_delayTimeR = _delayR / _sampleRate;
	}
	public function getDelayFeedBackLevelL():Number { return _feedBackL; }
	public function getDelayFeedBackLevelR():Number { return _feedBackR; }
	public function setDelayFeedBackLevelL(val:Number):void { _feedBackL = val; }
	public function setDelayFeedBackLevelR(val:Number):void { _feedBackR = val; }
	public function getDelayTimeL():Number { return _delayTimeL; }
	public function getDelayTimeR():Number { return _delayTimeR; }
	public function setDelayTimeL(val:Number):Number
	{
		var _frame:int = val * _sampleRate;
		if (_frame > _ringBufferMask)_frame = _ringBufferMask;
		_delayL = _frame;
		_rpL = _wp + (_ringBufferMask - _delayL) & _ringBufferMask;
		return _delayTimeL = _frame / _sampleRate;
	}
	public function setDelayTimeR(val:Number):Number
	{
		var _frame:int = val * _sampleRate;
		if (_frame > _ringBufferMask)_frame = _ringBufferMask;
		_delayR = _frame;
		_rpR = _wp + (_ringBufferMask - _delayR) & _ringBufferMask;
		return _delayTimeR = _frame / _sampleRate;
	}
	public function getDelayFrameL():int { return _delayL; }
	public function getDelayFrameR():int { return _delayR; }
	public function setDelayFrameL(val:int):void
	{ 
		_delayL = val;
		if (_delayL > _ringBufferMask)_delayL = _ringBufferMask;
		_rpL = _wp + (_ringBufferMask - _delayL) & _ringBufferMask;
		_delayTimeL = _delayL / _sampleRate;
	}
	public function setDelayFrameR(val:int):void
	{
		_delayR = val;
		if (_delayR > _ringBufferMask)_delayR = _ringBufferMask;
		_rpR = _wp + (_ringBufferMask - _delayR) & _ringBufferMask;
		_delayTimeR = _delayR / _sampleRate;
	}
	public function setEffectParam(mix:Number = 0.2, 
		delayTimeL:Number = 0.25, feedBackLevelL:Number = 0.4, 
		delayTimeR:Number = 0.35, feedBackLevelR:Number = 0.4
		):void
	{
		_mix = mix;
		_antiMix = 1.0 - _mix;
		setDelayTimeL(delayTimeL);
		setDelayTimeR(delayTimeR);
		_feedBackL = feedBackLevelL;
		_feedBackR = feedBackLevelR;
	}
// Constructor
	public function EFXStereoDelay(bufferSizeBits:int = 15):void
	{
		super();
		// bufferSize
		if (bufferSizeBits > 20) bufferSizeBits = 20;
		else if (bufferSizeBits < 4) bufferSizeBits = 4;
		_ringBufferSizeBits = bufferSizeBits;
		_ringBufferSize = 1 << _ringBufferSizeBits;
		_ringBufferMask = _ringBufferSize - 1;
		_delayBufferL = new Vector.<Number>(_ringBufferSize, true);
		_delayBufferR = new Vector.<Number>(_ringBufferSize, true);
		// reset
		clearBuffer();
		setEffectParam();
	}
// Base - EffectorBase ---
	override public function applyEffect(outL:Vector.<Number>, outR:Vector.<Number>, startFrame:int, sampleFrames:int):void
	{
		var _delay:Number, i:int = startFrame,  
			_endFrame:int = i + sampleFrames;
		for ( ; i < _endFrame; i++,
			_rpL = _rpL + 1 & _ringBufferMask,
			_rpR = _rpR + 1 & _ringBufferMask,
			_wp  = _wp  + 1 & _ringBufferMask)
		{
		// Channel L --->
			_delay = _delayBufferL[_rpL];
			_delayBufferL[_wp] = outL[i] + _delay * _feedBackL;
			outL[i] = outL[i] * _antiMix + _delay * _mix;	// outout
		// Channel R --->
			_delay = _delayBufferR[_rpR];
			_delayBufferR[_wp] = outR[i] + _delay * _feedBackR;
			outR[i] = outR[i] * _antiMix + _delay * _mix;	// outout
		}
	}
	override public function resetParameters():void
	{
		setEffectParam();
	}
	override public function clearBuffer():void
	{
		for (var i:int = 0; i < _ringBufferSize; i++)_delayBufferL[i] = _delayBufferR[i] = 0.0;
	}
	override public function initialize():void
	{
		clearBuffer();
		setEffectParam();
	}
// Interface - IfGrayBoxModule -----
	public function updateSampleRate(sampleRate:Number):void{ setSampleRate(sampleRate); }
	public function updateAudioChannels(ch:int):void{}
	public function updateProcessBlockSize(size:int):void{}
	
	public function resetModule():void{ initialize(); }
	public function initiModule():void{ initialize(); }
	public function connectModule(host:GrayBox):void{ clearBuffer(); setSampleRate(host.getSampleRate()); }
	public function disconnectModule():void{ clearBuffer(); }
	public function updateBypassStatus(bypass:Boolean):void{ clearBuffer(); }
	
	public function startProcess():void{ clearBuffer(); }
	public function stopProcess():void{ clearBuffer(); }
	public function suspendProcess():void{ clearBuffer(); }
	public function resumeProcess():void{ clearBuffer(); }
	public function process(procData:ProcessData):void
	{
		// inline --- applyEffect()
		var i:int = procData.offsetFrames,  
			_endFrame:int = i + procData.sampleFrames,
			_delay:Number, 
			_outL:Vector.<Number> = procData.audioOut[0],
			_outR:Vector.<Number> = procData.audioOut[1];
		for ( ; i < _endFrame; i++,
			_rpL = _rpL + 1 & _ringBufferMask,
			_rpR = _rpR + 1 & _ringBufferMask,
			_wp  = _wp  + 1 & _ringBufferMask)
		{
		// Channel L ---------
			_delay = _delayBufferL[_rpL];
			_delayBufferL[_wp] = _outL[i] + _delay * _feedBackL;
			_outL[i] = _outL[i] * _antiMix + _delay * _mix;	// outout
		// Channel R ---------
			_delay = _delayBufferR[_rpR];
			_delayBufferR[_wp] = _outR[i] + _delay * _feedBackR;
			_outR[i] = _outR[i] * _antiMix + _delay * _mix;	// outout
		}
	}
}
}